#include "PerlinNoise.hpp"

Point vector_product(Point u, Point v) {
    Point u_v(0,0,0);
    u_v.x = u.y* v.z - u.z * v.y;
    u_v.y = u.z* v.x - u.x * v.z;
    u_v.z = u.x* v.y - u.y * v.x;
    return u_v;
}

PerlinNoise::PerlinNoise() : ImplicitSurface(Point(0,0,0)) {
    // permutation vector (source : http://mrl.nyu.edu/~perlin/noise/)
    p = {
		151,160,137,91,90,15,131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,
		8,99,37,240,21,10,23,190, 6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,
		35,11,32,57,177,33,88,237,149,56,87,174,20,125,136,171,168, 68,175,74,165,71,
		134,139,48,27,166,77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,
		55,46,245,40,244,102,143,54, 65,25,63,161,1,216,80,73,209,76,132,187,208, 89,
		18,169,200,196,135,130,116,188,159,86,164,100,109,198,173,186, 3,64,52,217,226,
		250,124,123,5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,
		189,28,42,223,183,170,213,119,248,152, 2,44,154,163, 70,221,153,101,155,167,
		43,172,9,129,22,39,253, 19,98,108,110,79,113,224,232,178,185, 112,104,218,246,
		97,228,251,34,242,193,238,210,144,12,191,179,162,241, 81,51,145,235,249,14,239,
		107,49,192,214, 31,181,199,106,157,184, 84,204,176,115,121,50,45,127, 4,150,254,
        138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180 };
    // Duplication of the permutation vector.
    p.insert(p.end(), p.begin(), p.end());
}

PerlinNoise::PerlinNoise(unsigned int seed) : ImplicitSurface(Point(0,0,0)) {
	p.resize(256);

	// Fill p with values from 0 to 255
	std::iota(p.begin(), p.end(), 0);

	// Initialize a random engine with seed
	std::default_random_engine engine(seed);

	// Shuffle  using the above random engine
	std::shuffle(p.begin(), p.end(), engine);

	// Duplicate the permutation vector
	p.insert(p.end(), p.begin(), p.end());
}

double PerlinNoise::operator() (double x, double y, double z) const {
    return z - noise(x, y, 0);
}

double PerlinNoise::noise(double _x, double _y, double _z) const {
    int X = (int) floor(_x) & 255;
    int Y = (int) floor(_y) & 255;
    int Z = (int) floor(_z) & 255;

    _x -= floor(_x);
    _y -= floor(_y);
    _z -= floor(_z);

    double u = fade(_x);
    double v = fade(_y);
    double w = fade(_z);

    int A = p[X] + Y;
    int AA = p[A] + Z;
    int AB = p[A + 1] + Z;
    int B = p[X + 1] + Y;
    int BA = p[B] + Z;
    int BB = p[B + 1] + Z;

    double res = lerp(
       w,
       lerp(
           v,
           lerp(u, grad(p[AA], _x, _y, _z), grad(p[BA], _x-1, _y, _z)),
           lerp(u, grad(p[AB], _x, _y-1, _z), grad(p[BB], _x-1, _y-1, _z))
       ),
       lerp(
           v,
           lerp(
               u,
               grad(p[AA+1], _x, _y, _z-1),
               grad(p[BA+1], _x-1, _y, _z-1)
           ),
           lerp(
               u,
               grad(p[AB+1], _x, _y-1, _z-1),
               grad(p[BB+1], _x-1, _y-1, _z-1)
           )
       )
    );
	return (res + 1.0)/2.0;
}

double PerlinNoise::noise(double _x, double _y) const {
    return noise(_x, _y, 0);
}

double PerlinNoise::fade(double t) const {
    return t * t * t * (t * (t * 6 - 15) + 10);
}

double PerlinNoise::lerp(double t, double a, double b) const {
    return a + t * (b-a);
}

double PerlinNoise::grad(int hash, double x, double y, double z)  const {
	int h = hash & 15;
	// Convert lower 4 bits of hash into 12 gradient directions
	double u = h < 8 ? x : y,
		   v = h < 4 ? y : h == 12 || h == 14 ? x : z;
	return ((h & 1) == 0 ? u : -u) + ((h & 2) == 0 ? v : -v);
}

Point PerlinNoise::normal_vector(double x, double y) {
    double n_x_y = noise(x, y, 0);
    double n_dx_y = noise(x + 0.01, y, 0);
    double n_x_dy = noise(x, y + 0.01, 0);
    Point product = vector_product(
       Point(0.01, 0, n_dx_y - n_x_y),
       Point(0, 0.01, (n_x_dy - n_x_y))
    );
    return product;
}

Point PerlinNoise::location_hint() const {
    return location_hint(0, 0);
}

Point PerlinNoise::location_hint(double x, double z) const {
    return Point(x, noise(x, z), z);
}
